/* choice -- match incoming list against a collection of stored templates. */

/*  Copyright 1999 Miller Puckette.
Permission is granted to use this software for any purpose provided you
keep this copyright notice intact.

THE AUTHOR AND HIS EMPLOYERS MAKE NO WARRANTY, EXPRESS OR IMPLIED,
IN CONNECTION WITH THIS SOFTWARE.
 
This file is downloadable from msp.ucsd.edu .
*/

#include "m_pd.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
static t_class *kmeans_class;


#define DIM 10.
#define M 5.
#define C1 20
#define C2 20
#define C3 20
#define MAXCLUSTER 100

typedef struct cluster
{
    float c_x;
    float c_y;
    float c_xx;
    float c_xy;
    float c_yy;
} t_cluster;

static t_cluster kmeans_clusters[] =
{
    {3, 3, 0.5, 0, 0.5},
    {6, 3, 1, 0.7, 1},
    {2, 7, 1, -0.7, 1},
};

#define NPTS ((int)(sizeof(kmeans_clusters) / sizeof(*kmeans_clusters))) * C1
typedef struct _pt
{
    float p_x;
    float p_y;
} t_pt;

typedef struct _kmeans
{
    t_object x_obj;
    t_outlet *x_out1;
    t_outlet *x_out2;
    t_pt x_pt[NPTS];
    int x_n;
    int x_ncluster;
    int x_which[NPTS];
    t_pt x_means[MAXCLUSTER];
} t_kmeans;

int randi(int range)
{
    int n = (int)((double)(random()) *
        (double)(range) / RAND_MAX);
    if (n >= range)
        n = range;
    return (n);
}

    /* dumb gaussian */
float gauss(void)
{
    return (((randi(1000) + randi(1000) + randi(1000) +
    	randi(1000) + randi(1000)) - 2500) / 1000.);
}

static void kmeans_newpoints(t_kmeans *x)
{
    int i, j;
    for (i = 0; i < (int)(sizeof(kmeans_clusters)/sizeof(*kmeans_clusters));
    	i++)
            for (j = 0; j < C1; j++)
    {
    	t_cluster *c = &kmeans_clusters[i];
    	float g1 = gauss();
	float g2 = gauss();
	x->x_pt[i*C1+j].p_x = c->c_x + c->c_xx * g1 + c->c_xy * g2;
	x->x_pt[i*C1+j].p_y = c->c_y + c->c_xy * g1 + c->c_yy * g2;
	x->x_which[i*C1+j] = randi(3);
    }
}

static void kmeans_start(t_kmeans *x)
{
    int i, j;
    for (i = 0; i < NPTS; i++)
    {
	x->x_which[i] = randi(3);
    }
}

static void kmeans_ncluster(t_kmeans *x, t_floatarg f)
{
    if (f < 1) f = 1;
    else if (f > MAXCLUSTER)
    	f = MAXCLUSTER;
    x->x_ncluster = f;
}

static void *kmeans_new(void)
{
    int i, j;
    t_kmeans *x = (t_kmeans *)pd_new(kmeans_class);
    x->x_out1 = outlet_new(&x->x_obj, gensym("list"));
    x->x_out2 = outlet_new(&x->x_obj, gensym("list"));
    x->x_ncluster = 3;
    kmeans_start(x);
    return (x);
}

static void kmeans_print(t_kmeans *x)
{
}

static void kcluster_getmeans(t_kmeans *x)
{
    int i, j;
    for (j = 0; j < x->x_ncluster; j++)
    {
        float sumx, sumy, n;
        for (i = 0, sumx = sumy = n = 0; i < NPTS; i++)
            if (x->x_which[i] == j)
                sumx += x->x_pt[i].p_x, sumy += x->x_pt[i].p_y, n += 1;
        if (n <= 0) 
            for (i = 0, sumx = sumy = n = 0; i < NPTS; i++)
                sumx += x->x_pt[i].p_x, sumy += x->x_pt[i].p_y, n += 1;
        x->x_means[j].p_x = sumx/n;
        x->x_means[j].p_y = sumy/n;
    }
}

static void kmeans_list(t_kmeans *x, t_symbol *s, int argc, t_atom *argv)
{
    int i, j, k, bestj;
    float closest;
    t_atom at[3];
    int ntrain = atom_getfloatarg(0, argc, argv);
    float param1 = atom_getfloatarg(1, argc, argv);
    float param2 = atom_getfloatarg(2, argc, argv);
    kcluster_getmeans(x);
    for (k = 0; k < ntrain; k++)
    {
        for (i = 0; i < NPTS; i++)
        {
            float px = x->x_pt[i].p_x;
            float py = x->x_pt[i].p_y;
            for (j = 0, closest = 1e20, bestj = 0; j < x->x_ncluster; j++)
            {
                float dist =
                    (x->x_means[j].p_x - px) * (x->x_means[j].p_x - px)
                    + (x->x_means[j].p_y - py) * (x->x_means[j].p_y - py);
                if (dist < closest)
                    closest = dist, bestj = j;
            }
            x->x_which[i] = bestj;
        }
        kcluster_getmeans(x);
    }
    for (j = 0; j < x->x_ncluster; j++)
    {
        SETFLOAT(at+0, x->x_means[j].p_x);
        SETFLOAT(at+1, x->x_means[j].p_y);
        outlet_list(x->x_out2, 0, 2, at);
    }
    for (i = 0; i < NPTS; i++)
    {
        SETFLOAT(at+0, x->x_pt[i].p_x);
        SETFLOAT(at+1, x->x_pt[i].p_y);
        SETFLOAT(at+2, x->x_which[i]);
        outlet_list(x->x_out1, 0, 3, at);
    }
}

#define DSQUARED(x1, y1, x2, y2) ((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1))

static void kmeans_qccluster(t_kmeans *x, t_floatarg diam)
{
 #if 0
    /* QC clustering; fill this in later */
   int i, j, k1, k2, besti, nfound, bestn;
    float d2 = diam * diam;
    for (i = 0; i < NPTS; i++)
        x->x_which[i] = 0;
    for (j = 1; j < x->x_ncluster; j++)
    {
        for (k1 = 0, besti = 0; k1 < NPTS; k1++)
        {
            if (x->x_which[k1])
                continue;
            for (nfound = k2 = 0; k2 < NPTS; k2++)
            {
                if (x->x_which[k2])
                    continue;
                if (DSQUARED(x->x_pt[k1].p_x, x->x_pt[k1].p_y, 
                    x->x_pt[k2].p_x, x->x_pt[k2].p_y) < d2)
                        nfound++;
            }
            if (nfound > besti)
            {
            
    }
#endif 
}

void kmeans_setup(void)
{
    kmeans_class = class_new(gensym("kmeans"), (t_newmethod)kmeans_new,
        0, sizeof(t_kmeans), 0, A_DEFFLOAT, 0);
    class_addmethod(kmeans_class, (t_method)kmeans_newpoints,
    	gensym("newpoints"), 0);
    class_addmethod(kmeans_class, (t_method)kmeans_ncluster,
    gensym("ncluster"), A_FLOAT, 0);
    class_addmethod(kmeans_class, (t_method)kmeans_start, gensym("start"), 0);
    class_addmethod(kmeans_class, (t_method)kmeans_print, gensym("print"), 0);
    class_addmethod(kmeans_class, (t_method)kmeans_qccluster, 
        gensym("qccluster"), A_FLOAT, 0);
    class_addlist(kmeans_class, kmeans_list);
}
